use smallvec::smallvec;

use crate::func::prelude::*;
use super::maps::ExtentMap;

function! {
    /// `box`: Layouts content into a box.
    #[derive(Debug, PartialEq)]
    pub struct Boxed {
        body: SyntaxTree,
        map: ExtentMap<PSize>,
        debug: bool,
    }

    parse(args, body, ctx) {
        Boxed {
            body: parse!(optional: body, ctx).unwrap_or(SyntaxTree::new()),
            map: ExtentMap::new(&mut args, false)?,
            debug: args.get_key_opt::<bool>("debug")?.unwrap_or(true),
        }
    }

    layout(self, mut ctx) {
        ctx.repeat = false;
        ctx.debug = self.debug;

        let map = self.map.dedup(ctx.axes)?;

        // Try to layout this box in all spaces.
        let mut error = None;
        for &space in &ctx.spaces {
            let mut ctx = ctx.clone();
            let mut space = space;

            for &axis in &[Horizontal, Vertical] {
                if let Some(psize) = map.get(axis) {
                    let size = psize.scaled(ctx.base.get(axis));
                    *ctx.base.get_mut(axis) = size;
                    *space.dimensions.get_mut(axis) = size;
                    *space.expansion.get_mut(axis) = true;
                }
            }

            ctx.spaces = smallvec![space];

            match layout(&self.body, ctx) {
                Ok(layouts) => return Ok(vec![AddMultiple(layouts)]),
                Err(err) => error = Some(err),
            }
        }

        return Err(error.expect("expected at least one space"));
    }
}
